#include <stdarg.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <string.h>

static const char *Opt_lcd = NULL;
static const char *Opt_crt = NULL;

struct i810_par {
  unsigned char *mmio;
  int mem;
  int flag;
};

#define i810_readb(mmio_base, where)                     \
        *((volatile char *) (mmio_base + where))           \

#define i810_readw(mmio_base, where)                     \
       *((volatile short *) (mmio_base + where))           \

#define i810_readl(mmio_base, where)                     \
       *((volatile int *) (mmio_base + where))           \

#define i810_writeb(mmio_base, where, val)                              \
	*((volatile char *) (mmio_base + where)) = (volatile char) val      \

#define i810_writew(mmio_base, where, val)                              \
	*((volatile short *) (mmio_base + where)) = (volatile short) val    \

#define i810_writel(mmio_base, where, val)                              \
	*((volatile int *) (mmio_base + where)) = (volatile int) val    \


#define I810_MEMOPEN 1
#define I810_MMIO    2

/* registers */
#define PWR_CLKC              0x06014
#define HVSYNC                0x05000 
#define LCDTV_C               0x60018

#define MAXLINE               2048
#define I810STR               "8086:7121"
#define I810ESTR              "8086:7123"
#define I810_DC100STR_1       "8086:7125"
#define I810_DC100STR_2       "8086:1102"
#define I810_IGSTR            "8086:1112"
#define I810_CFCSTR           "8086:1132"
#define MEMSTR                "Memory at"
#define NONPRSTR              "32-bit, non-prefetchable"

static void release_resource(struct i810_par *par)
{
  if (par->flag & I810_MEMOPEN)
    close(par->mem);
  if (par->flag & I810_MMIO)
    munmap(par->mmio, 512 * 1024);
}

int strindex(char s[], char t[])
{
  int i, j, k;

  for (i = 0; s[i] != '\0'; i++) {
    for (j = i, k = 0; t[k] != '\0' && s[j] == t[k]; j++, k++)
      ;
    if (k > 0 && t[k] == '\0')
      return i;
  }
  return -1;
}

int i810_chip(char *buff, FILE *pci_f)
{
  int i, j, len=MAXLINE;

  while (getline(&buff, &len, pci_f) > 0) {
    j = (i = strindex(buff, I810STR)) >= 0 ||
        (i = strindex(buff, I810ESTR)) >= 0 ||
        (i = strindex(buff, I810_DC100STR_1)) >= 0 ||
        (i = strindex(buff, I810_DC100STR_2)) >= 0 ||
        (i = strindex(buff, I810_IGSTR)) >= 0 ||
        (i = strindex(buff, I810_CFCSTR)) >= 0;
    if (j) return i;
  }
  return -1;
}
	
unsigned long i810_addr(char *buff, FILE *pci_f)
{
  int i, len = MAXLINE;

  while (getline(&buff, &len, pci_f) > 0)
    if (strindex(buff, NONPRSTR) >= 0) {
      i = strindex(buff, MEMSTR);
      if (i >= 0)
	return strtoul(buff+i+sizeof(MEMSTR), NULL, 16);
    }
  return 0;
}

int i810_usage ()
{
  printf("usage: i810switch [crt on/off] [lcd on/off]\n");
  printf("  crt: enables/disables the output to the CRT display\n");
  printf("  lcd: enables/disables the output to the LCD\n");
  printf("  no options: displays the current output status\n");
  return 1;
}

int main (int argc, char *argv[])
{
  struct i810_par par;
  unsigned long addr;
  int i, crt = -1, lcd = -1, err = 0, count = 0, chip;
  FILE *pci_f;
  char buff[MAXLINE];
  char lspcistr[22] = "lspci -v -d xxxx:xxxx";

  if (geteuid() != 0) {
    printf("i810switch  must be setuid root.\n");
    return 1;
  }

  while (--argc > 0) {
    argv++;
    if (!strcmp(argv[0], "crt")) {
      if (--argc > 0) {
	argv++;
	if (!strcmp(argv[0], "on"))
	  crt = 1;
	else if (!strcmp(argv[0], "off"))
	  crt = 0;
	else
	  return i810_usage();
      } else {
	return i810_usage();
      }
    } else if (!strcmp(argv[0], "lcd")) {
      argv++;
      if (--argc > 0) {
	if (!strcmp(argv[0], "on"))
	  lcd = 1;
	else if (!strcmp(argv[0], "off"))
	  lcd = 0;
	else
	  return i810_usage();
      } else {
        return i810_usage();
      }
    } else {
      return i810_usage();
    }
  }

  pci_f = popen("lspci -n", "r");
  if (!pci_f) {
    printf("Something is wrong with lspci.\n");
    return 1;
  }
  chip = i810_chip(buff, pci_f);
  if (chip < 0) {
    printf("PCI id of i810 is not a recognized.\n");
    return 1;
  }
  pclose(pci_f);

  memcpy(lspcistr + 12, buff + chip, 9);
  pci_f = popen(lspcistr, "r");
  if (!pci_f) {
    printf("Something is wrong with lspci.\n");
    return 1;
  }
  addr = i810_addr(buff, pci_f);
  if (addr == 0) {
    printf("Something is wrong with lspci.\n");
    return 1;
  }
  pclose(pci_f);

  memset(&par, 0, sizeof(struct i810_par));

  par.mem = open("/dev/mem", O_RDWR);
  if (par.mem == -1) {
    printf("cannot open /dev/mem\n");
    release_resource(&par);
    return 1;
  }
  par.flag |= I810_MEMOPEN;

  par.mmio = mmap(NULL, 512 * 1024, PROT_WRITE | PROT_READ, 
		  MAP_SHARED, par.mem, addr);
  if (!par.mmio) {
    release_resource(&par);
    return 1;
  }

  par.flag |= I810_MMIO;

  if ((crt == -1) && (lcd == -1)) {
    printf("CRT: ");
    if (i810_readl(par.mmio, PWR_CLKC) & 1)
      printf("on\n");
    else
      printf("off\n");

    printf("LCD: ");
    if (i810_readl(par.mmio, LCDTV_C) & 0xc00)
      printf("off\n");
    else
      printf("on\n");
  }
  
  if (crt != -1) {
    int pwr_clkc = i810_readl(par.mmio, PWR_CLKC);
    int hvsync = i810_readl(par.mmio, HVSYNC);
    if (crt == 0) {
      printf("Disabling CRT display...\n");
      pwr_clkc &= ~1;
      hvsync = 0xa0000;
    } else {
      printf("Enabling CRT display...\n");
      pwr_clkc |= 1;
      hvsync = 0;
    }
    i810_writel(par.mmio, PWR_CLKC, pwr_clkc);
    i810_writel(par.mmio, HVSYNC, hvsync);
  }
  
  if (lcd != -1) {
    int lcdtv_c = i810_readl(par.mmio, LCDTV_C);
    if (lcd == 0) {
      printf("Disabling LCD...\n");
      lcdtv_c |= 0xc00;
    } else {
      printf("Enabling LCD...\n");
      lcdtv_c &= ~(0xc00); 
    }
    i810_writel(par.mmio, LCDTV_C, lcdtv_c);
  }
      
  release_resource(&par);
  exit (0);
}

